# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
# Date: 13.11.2016                                                            #
# Author: Ole-Johan Skrede                                                    #
#                                                                             #
# Solution proposal as part of the exercise program in                        #
# INF4300 - Digital image analysis at the University of Oslo                  #
#                                                                             #
# ::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::: #
"""
Implementation of Skeleton by Zone of Influence (SKIZ), also known as the
generalized Voronoi diagram.

As part of the solution proposal for exercise 11, task 6.
"""

# pylint: disable=bad-indentation
# pylint: disable=import-error

import numpy as np
import scipy.ndimage as ndi
import skimage.morphology as morph

def one_morph_thickening(bin_image, s_hit, s_miss):
  """Perform a thinning of the binary input image"""
  hom = ndi.binary_hit_or_miss(bin_image, structure1=s_hit, structure2=s_miss)
  return bin_image | hom

def one_morph_thinning(bin_image, s_hit, s_miss):
  """Perform a thinning of the binary input image"""
  hom = ndi.binary_hit_or_miss(bin_image, structure1=s_hit, structure2=s_miss)
  return bin_image & (~hom)

def morph_thickening(bin_image, s_hit, s_miss):
  """Perform a thickening of the binary input by rotating the structure elements"""
  new_im = one_morph_thinning(bin_image, s_hit, s_miss)
  for k in range(1, 4):
    new_im = one_morph_thickening(new_im, np.rot90(s_hit, k), np.rot90(s_miss, k))
  return new_im

def morph_thinning(bin_image, s_hit, s_miss):
  """Perform a thinning of the binary input by rotating the structure elements"""
  new_im = one_morph_thinning(bin_image, s_hit, s_miss)
  for k in range(1, 4):
    new_im = one_morph_thinning(new_im, np.rot90(s_hit, k), np.rot90(s_miss, k))
  return new_im

def get_bg_skeleton(bin_image):
  """Get the skeleton of the background of the image by successive thickenings"""
  s1_hit = np.array([[0, 0, 0], [0, 0, 0], [1, 1, 1]])
  s1_miss = np.array([[1, 1, 1], [0, 1, 0], [0, 0, 0]])
  s2_hit = np.array([[0, 0, 0], [1, 0, 0], [1, 1, 0]])
  s2_miss = np.array([[0, 1, 1], [0, 1, 1], [0, 0, 0]])

  old_im = bin_image.copy()

  while True:
    new_im = morph_thickening(old_im, s1_hit, s1_miss)
    new_im = morph_thickening(new_im, s2_hit, s2_miss)

    #print(np.linalg.norm(old_im*1 - new_im*1))
    no_change = np.allclose(old_im, new_im)
    if no_change:
      break
    else:
      old_im = new_im

  return new_im

def prune_bg_skeleton(skeleton):
  """Remove spurious edges of a background skeleton by successive thickenings"""
  s1_hit = np.array([[1, 0, 0], [1, 0, 1], [1, 1, 1]])
  s1_miss = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]])
  s2_hit = np.fliplr(s1_hit)
  s2_miss = np.fliplr(s1_miss)

  old_im = skeleton.copy()

  while True:
    new_im = morph_thickening(old_im, s1_hit, s1_miss)
    new_im = morph_thickening(new_im, s2_hit, s2_miss)

    #print(np.linalg.norm(old_im*1 - new_im*1))
    no_change = np.allclose(old_im, new_im)
    if no_change:
      break
    else:
      old_im = new_im

  return new_im

def get_skeleton(bin_image):
  """Get the skeleton of the image by successive thickenings"""
  s1_hit = np.array([[0, 0, 0], [0, 1, 0], [1, 1, 1]])
  s1_miss = np.array([[1, 1, 1], [0, 0, 0], [0, 0, 0]])
  s2_hit = np.array([[0, 0, 0], [1, 1, 0], [1, 1, 0]])
  s2_miss = np.array([[0, 1, 1], [0, 0, 1], [0, 0, 0]])

  old_im = bin_image.copy()

  while True:
    new_im = morph_thinning(old_im, s1_hit, s1_miss)
    new_im = morph_thinning(new_im, s2_hit, s2_miss)

    #print(np.linalg.norm(old_im*1 - new_im*1))
    no_change = np.allclose(old_im, new_im)
    if no_change:
      break
    else:
      old_im = new_im

  return new_im

def prune_skeleton(skeleton):
  """Remove spurious edges of a skeleton by successive thinnings"""
  s1_hit = np.array([[0, 0, 0], [0, 1, 0], [0, 0, 0]])
  s1_miss = np.array([[1, 0, 0], [1, 0, 1], [1, 1, 1]])
  s2_hit = np.fliplr(s1_hit)
  s2_miss = np.fliplr(s1_miss)

  old_im = skeleton.copy()

  while True:
    new_im = morph_thinning(old_im, s1_hit, s1_miss)
    new_im = morph_thinning(new_im, s2_hit, s2_miss)

    #print(np.linalg.norm(old_im*1 - new_im*1))
    no_change = np.allclose(old_im, new_im)
    if no_change:
      break
    else:
      old_im = new_im

  return new_im

def skiz(bin_image):
  """Compute the skeleton by zone of influence (SKIZ) of a binary image.

  Essentially, what we are doing is finding the skeleton of the background,
  and then remove "dead ends" (pruning) from the skeleton.

  Note 1:
    It seems like finding the skeleton or the medial axis give essentially the
    same result.

  Note 2:
    If we only do this once, the skeleton will leave small loops which is
    (correctly) not removed by the pruning. Therefore we fill these small holes,
    and do the procedure again. This we continue to do until convergence.

  Note 3:
    The pruning will interpret (correctly) a "grain" connected to the
    boundary as a "dead end". This is not desirable for us, therefore we pad
    the image with a frame of size 1 filled with ones.
  """
  if bin_image.dtype is not bool:
    bin_image = bin_image > 0

  #bg_skeleton = get_bg_skeleton(bin_image)
  #pruned = prune_bg_skeleton(bg_skeleton)


  #bg_skeleton = get_skeleton(~bin_image)
  bg_skeleton = morph.medial_axis(~bin_image)
  bg_skeleton = np.pad(bg_skeleton, 1, 'constant', constant_values=1).astype(bool)
  pruned = prune_skeleton(bg_skeleton)
  pruned = pruned[1:-1, 1:-1]

  # As is evident if we just prune the skeleton once, there are still somee
  # small loops which will cause the pruning to not be satisfactory. Albeit a
  # bit hacky, we fill these small holes until the skeleton is stabilized
  # before we prune.

  hole_size = 1000
  max_iter = 20
  niter = 0

  while niter < max_iter:
    small_filled = morph.remove_small_holes(pruned, hole_size)
    no_change = np.allclose(pruned, small_filled)
    if no_change:
      break
    else:
      bg_skeleton = morph.medial_axis(small_filled)
      bg_skeleton = np.pad(bg_skeleton, 1, 'constant', constant_values=1).astype(bool)
      pruned = prune_skeleton(bg_skeleton)
      pruned = pruned[1:-1, 1:-1]
    niter += 1

  return pruned
